"""
WOfS (wofl) product specification
=================================

Each value in a ``wofl`` indicates whether it contains a valid water determination,
and if it is not valid, why it has been excluded.

The clear and valid observations are:

clear dry == 0

clear wet == 128

===  =============  ==========  =======
Bit  Decimal value  Value       Meaning
===  =============  ==========  =======
     0              0           no water present
0    1              1<<0        nodata (missing all earth observation bands)
1    2              1<<1        noncontiguous (at least one EO band is missing or saturated)
2    4              1<<2        low solar incidence angle
3    8              1<<3        terrain shadow
4    16             1<<4        high slope
5    32             1<<5        cloud shadow
6    64             1<<6        cloud
7    128            1<<7        classified as water by the decision tree
===  =============  ==========  =======

The land/sea mask (bit 2) should be ignored. It is based on a vector sea mask
which excludes useful data. We are interested in keeping ocean observations
anyway.
"""

# pylint: disable=bad-whitespace, line-too-long
# For the following bits, 0=unmasked
WATER_PRESENT = 1 << 7  # (dec 128) bit 7: 1=water present, 0=no water if all other bits zero
MASKED_CLOUD = 1 << 6  # (dec 64)  bit 6: 1=pixel masked out due to cloud
MASKED_CLOUD_SHADOW = 1 << 5  # (dec 32)  bit 5: 1=pixel masked out due to cloud shadow
MASKED_HIGH_SLOPE = 1 << 4  # (dec 16)  bit 4: 1=pixel masked out due to high slope
MASKED_TERRAIN_SHADOW = 1 << 3  # (dec 8)   bit 3: 1=pixel masked out due to terrain shadow
MASKED_LOW_SOLAR_ANGLE = 1 << 2  # (dec 4) bit 2: 1=pixel masked out due to low solar incidence angle
MASKED_NO_CONTIGUITY = 1 << 1  # (dec 2)   bit 1: 1=pixel masked out due to lack of data contiguity
NO_DATA = 1 << 0  # (dec 1)   bit 0: 1=pixel masked out due to NO_DATA in NBAR source, 0=valid data in NBAR
WATER_NOT_PRESENT = 0  # (dec 0)          All bits zero indicated valid observation, no water present

# Water detected on slopes equal or greater than this value are masked out
SLOPE_THRESHOLD_DEGREES = 12.0

# If the sun only grazes a hillface, observation unreliable (vegetation shadows etc)
LOW_SOLAR_INCIDENCE_THRESHOLD_DEGREES = 10



"""
Produce water observation feature layers (i.e. water extent foliation).

These are the WOfS product with temporal extent (i.e. multiple time values).
Consists of wet/dry estimates and filtering flags,
with one-to-one correspondence to earth observation layers.

Not to be confused with the wofs summary products,
which are derived from a condensed mosaic of the wofl archive.

Issues:
    - previous documentation may be ambiguous or previous implementations may differ
      (e.g. saturation, bitfield)
    - Tile edge artifacts concerning cloud buffers and cloud or terrain shadows.
    - DSM may have different natural resolution to EO source.
      Should think about what CRS to compute in, and what resampling methods to use.
      Also, should quantify whether earth's curvature is significant on tile scale.
    - Yet to profile memory, CPU or IO usage.
"""
import numpy as np

from wofs import classifier, filters
from wofs.constants import NO_DATA
from wofs.filters import eo_filter, fmask_filter, terrain_filter, c2_filter


def woffles(nbar, pq, dsm, dsm_no_data=-1000, ignore_dsm_no_data=False):
    """Generate a Water Observation Feature Layer from NBAR, PQ and surface elevation inputs."""

    water = classifier.classify(nbar.to_array(dim='band')) \
        | filters.eo_filter(nbar) \
        | filters.pq_filter(pq.pqa) \
        | filters.terrain_filter(
            dsm,
            nbar,
            no_data=dsm_no_data,
            ignore_dsm_no_data=ignore_dsm_no_data)

    _fix_nodata_to_single_value(water)

    assert water.dtype == np.uint8

    return water


def woffles_ard(ard, dsm, dsm_no_data=-1000, ignore_dsm_no_data=False):
    """Generate a Water Observation Feature Layer from ARD (NBART and FMASK) and surface elevation inputs."""
    nbar_bands = spectral_bands(ard)
    water = classifier.classify(nbar_bands) \
        | eo_filter(ard) \
        | fmask_filter(ard.fmask)

    if dsm is not None:
        # terrain_filter arbitrarily expects a band named 'blue'
        water |= terrain_filter(
            dsm,
            ard.rename({"nbart_blue": "blue"}),
            no_data=dsm_no_data,
            ignore_dsm_no_data=ignore_dsm_no_data
        )

    _fix_nodata_to_single_value(water)

    assert water.dtype == np.uint8

    return water


def woffles_usgs_c2(c2, dsm, dsm_no_data=-1000, ignore_dsm_no_data=False):
    """Generate a Water Observation Feature Layer from USGS Collection 2 and surface elevation inputs."""
    nbar_bands = spectral_bands(c2)
    water = classifier.classify(nbar_bands) \
        | eo_filter(c2) \
        | c2_filter(c2.fmask)
    if dsm is not None:
        # terrain_filter arbitrarily expects a band named 'blue'
        water |= terrain_filter(
            dsm,
            c2.rename({"nbart_blue": "blue"}),
            no_data=dsm_no_data,
            ignore_dsm_no_data=ignore_dsm_no_data
        )

    _fix_nodata_to_single_value(water)

    assert water.dtype == np.uint8

    return water


def spectral_bands(ds):
    bands = [
        "nbart_blue",
        "nbart_green",
        "nbart_red",
        "nbart_nir",
        "nbart_swir_1",
        "nbart_swir_2",
    ]
    return ds[bands].to_array(dim="band")


def _fix_nodata_to_single_value(dataarray):
    # Force any values with the NODATA bit set, to be the nodata value
    nodata_set = np.bitwise_and(dataarray.data, NO_DATA) == NO_DATA

    # If we don't specifically set the dtype in the following line,
    # dask arrays explode to int64s. Make sure it stays a uint8!
    dataarray.data[nodata_set] = np.array(NO_DATA, dtype="uint8")



"""
Entry point for producing WOfS products.

Specifically intended for running in the PBS job queue system at the NCI.

The three entry points are:
1. datacube-wofs submit
2. datacube-wofs generate
3. datacube-wofs run
"""
import copy
import logging
import os
import pickle
import signal
import sys
from collections import defaultdict
from copy import deepcopy
from datetime import datetime, timezone
from functools import partial
from pathlib import Path
from time import time as time_now
from typing import Tuple

import click
import datacube
import datacube.model.utils
import numpy as np
import xarray
from datacube.api.grid_workflow import Tile
from datacube.drivers.netcdf import write_dataset_to_netcdf
from datacube.index import Index, MissingRecordError
from datacube.model import DatasetType, Range
from datacube.ui import click as ui
from datacube.ui import task_app
from datacube.utils.geometry import unary_union, unary_intersection, CRS
from digitalearthau import paths
from digitalearthau.qsub import with_qsub_runner, TaskRunner
from digitalearthau.runners.model import TaskDescription
from pandas import to_datetime
from wofs import wofls, __version__

APP_NAME = 'wofs'
_LOG = logging.getLogger(__name__)
_MEASUREMENT_KEYS_TO_COPY = ('zlib', 'complevel', 'shuffle', 'fletcher32', 'contiguous', 'attrs')

# ROOT_DIR is the current directory of this file.
ROOT_DIR = Path(__file__).absolute().parent.parent

INPUT_SOURCES = [{'nbart': 'ls5_nbart_albers',
                  'pq': 'ls5_pq_legacy_scene',
                  'sensor_name': 'TM',
                  'platform_name': 'LANDSAT-5',
                  'platform_name_short': 'ls5',
                  'source_product': 'ls5_level1_scene'},
                 {'nbart': 'ls7_nbart_albers',
                  'pq': 'ls7_pq_legacy_scene',
                  'sensor_name': 'ETM',
                  'platform_name': 'LANDSAT-7',
                  'platform_name_short': 'ls7',
                  'source_product': 'ls7_level1_scene'},
                 {'nbart': 'ls8_nbart_albers',
                  'pq': 'ls8_pq_legacy_scene',
                  'sensor_name': 'OLI',
                  'platform_name': 'LANDSAT-8',
                  'platform_name_short': 'ls8',
                  'source_product': 'ls8_level1_scene'},
                 ]


def _make_wofs_config(index, config, dry_run):
    """
    Refine the configuration

    The task-app machinery loads a config file, from a path specified on the
    command line, into a dict. This function is an opportunity, with access to
    the datacube index, to modify that dict before it is used (being passed on
    to both the make-tasks and the do-task).

    For a dry run, we still need to create a dummy DatasetType to
    generate tasks (e.g. via the GridSpec), but a normal run must index
    it as a product in the database and replace the dummy with a fully-fleshed
    DatasetType since the tasks involve writing metadata to file that
    is specific to the database instance (note, this may change in future).
    """

    if not dry_run:
        _LOG.info('Created DatasetType %s', config['product_definition']['name'])  # true? not yet.

    config['wofs_dataset_type'] = _get_product(index, config['product_definition'], dry_run)

    config['variable_params'] = _build_variable_params(config)

    if 'task_timestamp' not in config:
        config['task_timestamp'] = int(time_now())

    if not os.access(config['location'], os.W_OK):
        _LOG.error('Current user appears not have write access output location: %s', config['location'])

    return config


def _build_variable_params(config: dict) -> dict:
    chunking = config['product_definition']['storage']['chunking']
    chunking = [chunking[dim] for dim in config['product_definition']['storage']['dimension_order']]

    variable_params = {}
    for mapping in config['product_definition']['measurements']:
        measurement_name = mapping['name']
        variable_params[measurement_name] = {
            k: v
            for k, v in mapping.items()
            if k in _MEASUREMENT_KEYS_TO_COPY
        }
        variable_params[measurement_name]['chunksizes'] = chunking
    return variable_params


def _create_output_definition(config: dict, source_product: DatasetType) -> dict:
    output_product_definition = deepcopy(source_product.definition)
    output_product_definition['name'] = config['product_definition']['name']
    output_product_definition['description'] = config['product_definition']['description']
    output_product_definition['managed'] = True
    var_def_keys = {'name', 'dtype', 'nodata', 'units', 'flags_definition'}

    output_product_definition['measurements'] = [
        {k: v for k, v in measurement.items() if k in var_def_keys}
        for measurement in config['product_definition']['measurements']
    ]

    output_product_definition['metadata']['format'] = {'name': 'NetCDF'}
    output_product_definition['metadata']['product_type'] = config.get('product_type', 'wofs')
    output_product_definition['metadata_type'] = config['product_definition']['metadata_type']

    output_product_definition['storage'] = {
        k: v for (k, v) in config['product_definition']['storage'].items()
        if k in ('crs', 'chunking', 'tile_size', 'resolution', 'dimension_order', 'driver')
    }

    # Validate the output product definition
    DatasetType.validate(output_product_definition)
    return output_product_definition


def _ensure_products(app_config: dict, index: Index, dry_run: bool, input_source) -> DatasetType:
    source_product_name = input_source
    source_product = index.products.get_by_name(source_product_name)
    if not source_product:
        raise ValueError(f"Source product {source_product_name} does not exist")

    output_product = DatasetType(
        source_product.metadata_type,
        _create_output_definition(app_config, source_product)
    )

    if not dry_run:
        _LOG.info('Add the output product definition for %s in the database.', output_product.name)
        output_product = index.products.add(output_product)

    return output_product


def _get_product(index, definition, dry_run):
    """
    Get the database record corresponding to the given product definition
    """
    metadata_type = index.metadata_types.get_by_name(definition['metadata_type'])
    prototype = DatasetType(metadata_type, definition)

    if not dry_run:
        _LOG.info('Add product definition to the database.')
        prototype = index.products.add(prototype)  # idempotent operation

    return prototype


def _get_filename(config, x, y, t):
    """
    Get file path from the config location
    """
    destination = config['location']
    filename_template = config['file_path_template']

    filename = filename_template.format(tile_index=(x, y),
                                        start_time=to_datetime(t).strftime('%Y%m%d%H%M%S%f'),
                                        version=config['task_timestamp'])  # A per JOB timestamp, seconds since epoch
    return Path(destination, filename)


def _group_tiles_by_cells(tile_index_list, cell_index_list):
    """
    Group the tiles by cells
    """
    key_map = defaultdict(list)
    for x, y, t in tile_index_list:
        if (x, y) in cell_index_list:
            key_map[(x, y)].append((x, y, t))
    return key_map


# pylint: disable=too-many-locals
def _generate_tasks(index, config, time, extent=None):
    """
    Yield tasks (loadables (nbart,ps,dsm) + output targets), for dispatch to workers.

    This function is the equivalent of an SQL join query,
    and is required as a workaround for datacube API abstraction layering.
    """
    extent = extent if extent is not None else {}
    product = config['wofs_dataset_type']

    assert product.grid_spec.crs == CRS('EPSG:3577')
    assert all((abs(r) == 25) for r in product.grid_spec.resolution)  # ensure approx. 25 metre raster
    pq_padding = [3 * 25] * 2  # for 3 pixel cloud dilation
    terrain_padding = [6850] * 2
    # Worst case shadow: max prominence (Kosciuszko) at lowest solar declination (min incidence minus slope threshold)
    # with snapping to pixel edges to avoid API questions
    # e.g. 2230 metres / math.tan(math.radians(30-12)) // 25 * 25 == 6850

    gw = datacube.api.GridWorkflow(index, grid_spec=product.grid_spec)  # GridSpec from product definition

    wofls_loadables = gw.list_tiles(product=product.name, time=time, **extent)
    dsm_loadables = gw.list_cells(product='dsm1sv10', tile_buffer=terrain_padding, **extent)

    if dsm_loadables:
        _LOG.info('Found %d dsm loadables', len(dsm_loadables))
    else:
        _LOG.warning('No dsm1sv10 product in the database')

    for input_source in INPUT_SOURCES:
        gqa_filter = dict(product=input_source['source_product'], time=time, gqa_iterative_mean_xy=(0, 1))
        nbart_loadables = gw.list_tiles(product=input_source['nbart'], time=time, **extent)
        pq_loadables = gw.list_tiles(product=input_source['pq'], time=time, tile_buffer=pq_padding, **extent)

        _LOG.info('Found %d nbart loadables for %r input source', len(nbart_loadables), input_source['nbart'])
        _LOG.info('Found %d pq  loadables for %r input source', len(pq_loadables), input_source['pq'])

        # only valid where EO, PQ and DSM are *all* available (and WOFL isn't yet)
        tile_index_set = (set(nbart_loadables) & set(pq_loadables)) - set(wofls_loadables)
        key_map = _group_tiles_by_cells(tile_index_set, dsm_loadables)

        _LOG.info('Found %d items for %r input source', len(list(key_map.keys())), input_source['source_product'])

        # Cell index is X,Y, tile_index is X,Y,T
        for cell_index, tile_indexes in key_map.items():
            geobox = gw.grid_spec.tile_geobox(cell_index)
            dsm_tile = gw.update_tile_lineage(dsm_loadables[cell_index])
            for tile_index in tile_indexes:
                nbart_tile = gw.update_tile_lineage(nbart_loadables.pop(tile_index))
                pq_tile = gw.update_tile_lineage(pq_loadables.pop(tile_index))
                valid_region = _find_valid_data_region(geobox, nbart_tile, pq_tile, dsm_tile)
                if not valid_region.is_empty:
                    yield dict(source_tile=nbart_tile,
                               pq_tile=pq_tile,
                               dsm_tile=dsm_tile,
                               file_path=_get_filename(config, *tile_index),
                               tile_index=tile_index,
                               extra_global_attributes=dict(platform=input_source['platform_name'],
                                                            instrument=input_source['sensor_name']),
                               valid_region=valid_region)


def _make_wofs_tasks(index, config, year=None, **kwargs):
    """
    Generate an iterable of 'tasks', matching the provided filter parameters.
    Tasks can be generated for:
     - all of time
     - 1 particular year
     - a range of years
    Tasks can also be restricted to a given spatial region, specified in `kwargs['x']` and `kwargs['y']` in `EPSG:3577`.
    """
    # TODO: Filter query to valid options
    if isinstance(year, int):
        query_time = Range(datetime(year=year, month=1, day=1), datetime(year=year + 1, month=1, day=1))
    elif isinstance(year, tuple):
        query_time = Range(datetime(year=year[0], month=1, day=1), datetime(year=year[1] + 1, month=1, day=1))
    else:
        query_time = year

    extent = {}
    if 'x' in kwargs and kwargs['x']:
        extent['crs'] = 'EPSG:3577'
        extent['x'] = kwargs['x']
        extent['y'] = kwargs['y']

    tasks = _generate_tasks(index, config, time=query_time, extent=extent)
    return tasks


def _get_app_metadata(config):
    """
    Get WOfS app metadata
    """
    doc = {
        'lineage': {
            'algorithm': {
                'name': APP_NAME,
                'version': __version__,
                'repo_url': 'https://github.com/GeoscienceAustralia/wofs.git',
                'parameters': {'configuration_file': str(config['app_config_file'])}
            },
        }
    }
    return doc


def _find_valid_data_region(geobox, *sources_list):
    """
    Find the valid data region
    """
    # perform work in CRS of the output tile geobox
    unfused = [[dataset.extent.to_crs(geobox.crs) for dataset in tile.sources.item()]
               for tile in sources_list]
    # fuse the dataset extents within each source tile
    # pylint: disable=map-builtin-not-iterating
    tiles_extents = map(unary_union, unfused)
    # find where (within the output tile) that all prerequisite inputs available
    return unary_intersection([geobox.extent] + list(tiles_extents))
    # downstream should check if this is empty..


def _docvariable(agdc_dataset, time):
    """
    Convert datacube dataset to xarray/NetCDF variable
    """
    array = xarray.DataArray([agdc_dataset], coords=[time])
    docarray = datacube.model.utils.datasets_to_doc(array)
    docarray.attrs['units'] = '1'  # unitless (convention)
    return docarray


def _do_wofs_task(config, task):
    """
    Load data, run WOFS algorithm, attach metadata, and write output.
    :param dict config: Config object
    :param dict task: Dictionary of task values

    :return: Dataset objects representing the generated data that can be added to the index
    :rtype: list(datacube.model.Dataset)
    """
    # datacube.api.Tile source_tile: NBAR Tile
    source_tile: Tile = task['source_tile']

    # datacube.api.Tile pq_tile: Pixel quality Tile
    pq_tile: Tile = task['pq_tile']

    # datacube.api.Tile dsm_tile: Digital Surface Model Tile
    dsm_tile: Tile = task['dsm_tile']

    # Path file_path: output file destination
    file_path = Path(task['file_path'])  # Path file_path: output file destination

    product = config['wofs_dataset_type']

    if file_path.exists():
        _LOG.warning('Output file already exists %r', str(file_path))

    # load data
    bands = ['blue', 'green', 'red', 'nir', 'swir1', 'swir2']  # inputs needed from EO data)
    source = datacube.api.GridWorkflow.load(source_tile, measurements=bands)
    pq = datacube.api.GridWorkflow.load(pq_tile)
    dsm = datacube.api.GridWorkflow.load(dsm_tile, resampling='cubic')

    # Core computation
    result = wofls.woffles(*(x.isel(time=0) for x in [source, pq, dsm])).astype(np.int16)

    # Convert 2D DataArray to 3D DataSet
    result = xarray.concat([result], dim=source.time).to_dataset(name='water')

    # add metadata
    result.water.attrs['nodata'] = 1  # lest it default to zero (i.e. clear dry)
    result.water.attrs['units'] = '1'  # unitless (convention)
    result.water.attrs['crs'] = source.crs

    # Attach CRS. Note this is poorly represented in NetCDF-CF
    # (and unrecognised in xarray), likely improved by datacube-API model.
    result.attrs['crs'] = source.crs

    # Provenance tracking
    parent_sources = [ds for tile in [source_tile, pq_tile, dsm_tile] for ds in tile.sources.values[0]]

    # Create indexable record
    new_record = datacube.model.utils.make_dataset(
        product=product,
        sources=parent_sources,
        center_time=result.time.values[0],
        uri=file_path.absolute().as_uri(),
        extent=source_tile.geobox.extent,
        valid_data=task['valid_region'],
        app_info=_get_app_metadata(config)
    )

    def harvest(what, tile):
        """
        Inherit optional metadata from EO, for future convenience only
        """
        datasets = [ds for source_datasets in tile.sources.values for ds in source_datasets]
        values = [dataset.metadata_doc[what] for dataset in datasets]
        assert all(value == values[0] for value in values)
        return copy.deepcopy(values[0])

    new_record.metadata_doc['platform'] = harvest('platform', source_tile)
    new_record.metadata_doc['instrument'] = harvest('instrument', source_tile)

    # copy metadata record into xarray
    result['dataset'] = _docvariable(new_record, result.time)

    global_attributes = config['global_attributes'].copy()
    global_attributes.update(task['extra_global_attributes'])

    # write output
    write_dataset_to_netcdf(result, file_path,
                            global_attributes=global_attributes,
                            variable_params=config['variable_params'])
    return [new_record]


def _index_datasets(index: Index, results):
    """
    Index newly created WOfS datasets
    """
    for dataset in results:
        try:
            index.datasets.add(dataset,
                               with_lineage=False,
                               sources_policy='skip')
            _LOG.info('Dataset %s added: %r', dataset.id, dataset)
        except (ValueError, MissingRecordError) as err:
            _LOG.error('Failed to add %r dataset: Error (%s)',
                       dataset,
                       err)


def _skip_indexing_and_only_log(results):
    _LOG.info(f'Skipping Indexing {len(results)} datasets')

    for dataset in results:
        _LOG.info('Dataset %s created at %s but not indexed', dataset.id, dataset.uris)


def handle_sigterm(signum, frame):
    # TODO: I expect SIGTERM will be received shortly before PBS SIGKILL's the job.
    # We could use this to print out the current progress of the job if it's incomplete
    # Or perhaps even make it resumable.
    pid = os.getpid()
    _LOG.error(f'WOfS App PID: {pid}. Received SIGTERM')


@click.group(help='Datacube WOfS')
@click.version_option(version=__version__)
def cli():
    """
    Instantiate a click 'Datacube WOfS' group object to register the following sub-commands for
    different bits of WOfS processing:
         1) list
         2) ensure-products
         4) generate
         5) run
    :return: None
    """
    signal.signal(signal.SIGTERM, handle_sigterm)


@cli.command(name='list', help='List installed WOfS config files')
def list_configs():
    """
     List installed WOfS config files
    :return: None
    """
    # Since wofs config files are packaged two levels down the ROOT_DIR,
    # ROOT_DIR.parents[2] will ensure that we point to dea/<YYYYMMDD> directory.
    for cfg in ROOT_DIR.parents[2].glob('wofs/config/*.yaml'):
        click.echo(cfg)


@cli.command(name='ensure-products',
             help="Ensure the products exist for the given WOfS config, create them if necessary.")
@task_app.app_config_option
@click.option('--dry-run', is_flag=True, default=False,
              help='Check product definition without modifying the database')
@ui.config_option
@ui.verbose_option
@ui.pass_index(app_name=APP_NAME)
def ensure_products(index, app_config, dry_run):
    """
    Ensure the products exist for the given WOfS config, creating them if necessary.
    If dry run is disabled, the validated output product definition will be added to the database.
    """
    # TODO: Add more validation of config?
    click.secho(f"Loading {app_config}", bold=True)
    out_product = _ensure_products(paths.read_document(app_config),
                                   index,
                                   dry_run,
                                   'wofs_albers')
    click.secho(f"Output product definition for {out_product.name} product exits in the database for the given "
                f"WOfS input config file")


@cli.command(help='Generate Tasks into a queue file for later processing')
@click.option('--app-config', help='WOfS configuration file',
              required=True,
              type=click.Path(exists=True, readable=True, writable=False, dir_okay=False))
@click.option('--output-filename',
              help='Filename to write the list of tasks to.',
              required=True,
              type=click.Path(exists=False, writable=True, dir_okay=False))
@click.option('--year', 'time_range',
              type=int,
              help='Limit the process to a particular year, or "-" separated range of years.')
@click.option('--dry-run', is_flag=True, default=False,
              help='Check product definition without modifying the database')
@ui.verbose_option
@ui.log_queries_option
@ui.pass_index(app_name=APP_NAME)
def generate(index: Index,
             app_config: str,
             output_filename: str,
             dry_run: bool,
             time_range: Tuple[datetime, datetime]):
    """
    Generate Tasks into file and Queue PBS job to process them

    By default, also ensures the Output Product is present in the database.

    --dry-run will still generate a tasks file, but not add the output product to the database.
    """
    app_config_file = Path(app_config).resolve()
    app_config = paths.read_document(app_config_file)

    wofs_config = _make_wofs_config(index, app_config, dry_run)

    # Patch in config file location, for recording in dataset metadata
    wofs_config['app_config_file'] = app_config_file

    wofs_tasks = _make_wofs_tasks(index, wofs_config, time_range)
    num_tasks_saved = task_app.save_tasks(
        wofs_config,
        wofs_tasks,
        output_filename
    )
    _LOG.info('Found %d tasks', num_tasks_saved)


@cli.command(help="Display information about a tasks file")
@click.argument('task_file')
def inspect_taskfile(task_file):
    with open(task_file, 'rb') as fin:
        config = pickle.load(fin)
        print('CONFIGURATION')
        print(config)
        print('\nFIRST TASK')
        task1 = pickle.load(fin)
        print(task1)
        i = 1
        while True:
            try:
                _ = pickle.load(fin)
                i += 1
            except EOFError:
                break
        print(f'{i} tasks in total')


@cli.command(help='Check for existing outputs')
@click.option('--input-filename', required=True,
              help='A Tasks File to process',
              type=click.Path(exists=True, readable=True, writable=False, dir_okay=False))
def check_existing(input_filename: str):
    config, tasks = task_app.load_tasks(input_filename)

    _LOG.info('Checking for existing output files.')
    # tile_index is X, Y, T
    task_app.check_existing_files(_get_filename(config, *task['tile_index']) for task in tasks)


def _prepend_path_to_tasks(prepath, tasks):
    """
    Prepend prepath to each task's file_path

    Taking special care when prepending to absolute paths.
    """
    for task in tasks:
        file_path = task['file_path']
        if file_path.is_absolute():
            task['file_path'] = Path(prepath).joinpath(*file_path.parts[1:])
        else:
            task['file_path'] = Path(prepath) / file_path
        yield task


@cli.command(help='Process generated task file')
@click.option('--input-filename', required=True,
              help='A Tasks File to process',
              type=click.Path(exists=True, readable=True, writable=False, dir_okay=False))
@click.option('--skip-indexing', is_flag=True, default=False,
              help="Generate output files but don't record to a database index")
@click.option('--redirect-outputs',
              help='Store output files in a different directory (for testing, prepended to task defined output)',
              type=click.Path(exists=True, dir_okay=True, file_okay=False, writable=True))
@with_qsub_runner()
@ui.verbose_option
@ui.pass_index(app_name=APP_NAME)
def run(index,
        input_filename: str,
        runner: TaskRunner,
        skip_indexing: bool,
        redirect_outputs: str,
        **kwargs):
    """
    Process WOfS tasks from a task file.
    """
    config, tasks = task_app.load_tasks(input_filename)
    work_dir = Path(input_filename).parent

    if redirect_outputs is not None:
        tasks = _prepend_path_to_tasks(redirect_outputs, tasks)

    # TODO: Get rid of this completely
    task_desc = TaskDescription(
        type_='wofs',
        task_dt=datetime.utcnow().astimezone(timezone.utc),
        events_path=work_dir,
        logs_path=work_dir,
        jobs_path=work_dir,
        parameters=None,
        runtime_state=None,
    )

    _LOG.info('Starting WOfS processing...')
    task_func = partial(_do_wofs_task, config)
    if skip_indexing:
        process_func = _skip_indexing_and_only_log
    else:
        process_func = partial(_index_datasets, index)

    try:
        runner(task_desc, tasks, task_func, process_func)
        _LOG.info("Runner finished normally, triggering shutdown.")
    finally:
        runner.stop()

    # TODO: Check for failures and return error state
    sys.exit(0)


@cli.command(name='run-mpi', help='Run using MPI')
@click.option('--skip-indexing', is_flag=True, default=False,
              help="Generate output files but don't record to database")
@click.option('--input-filename', required=True,
              help='A Tasks File to process',
              type=click.Path(exists=True, readable=True, writable=False,
                              dir_okay=False))
@click.option('--redirect-outputs',
              help='Store output files in a different directory (for testing,'
                   'prepended to task defined output)',
              type=click.Path(exists=True, dir_okay=True, file_okay=False,
                              writable=True))
@ui.verbose_option
@ui.pass_index(app_name=APP_NAME)
def mpi_run(index,
            input_filename: str,
            skip_indexing: bool,
            redirect_outputs: str,
            **kwargs):
    """Process generated task file.

    Iterate over the file list and assign MPI worker for processing. Split the
    input file by the number of workers, each MPI worker completes every nth
    task. Also, detect and fail early if not using full resources in an MPI job.

    Before using this command, execute the following:
      $ module use /g/data/v10/public/modules/modulefiles/
      $ module load dea
      $ module load openmpi

    """
    config, tasks = task_app.load_tasks(input_filename)

    if redirect_outputs is not None:
        tasks = _prepend_path_to_tasks(redirect_outputs, tasks)

    _LOG.info('Starting WOfS processing...')
    if skip_indexing:
        process_func = _skip_indexing_and_only_log
    else:
        process_func = partial(_index_datasets, index)

    for task in _nth_by_mpi(tasks):
        try:
            dataset = _do_wofs_task(config, task)
            process_func(dataset)

            _LOG.info('Successfully processed', task)
        except Exception:
            _LOG.exception('Unable to process', task)


def _mpi_init():
    """Ensure we're running within a good MPI environment, and find out the number
    of processes we have.

    """
    from mpi4py import MPI
    job_rank = MPI.COMM_WORLD.rank  # Rank of this process
    job_size = MPI.COMM_WORLD.size  # Total number of processes
    universe_size = MPI.COMM_WORLD.Get_attr(MPI.UNIVERSE_SIZE)
    pbs_ncpus = os.environ.get('PBS_NCPUS', None)
    if pbs_ncpus is not None and int(universe_size) != int(pbs_ncpus):
        _LOG.error('Running within PBS and not using all available resources. Abort!')
        sys.exit(1)
    _LOG.info('MPI Info', mpi_job_size=job_size, mpi_universe_size=universe_size, pbs_ncpus=pbs_ncpus)
    return job_rank, job_size


def _nth_by_mpi(iterator):
    """
    Split an iterator across MPI processes

    Based on MPI pool size and rank of this process
    """
    from mpi4py import MPI
    job_size = MPI.COMM_WORLD.size  # Total number of processes
    job_rank = MPI.COMM_WORLD.rank  # Rank of this process
    for i, element in enumerate(iterator):
        if i % job_size == job_rank:
            yield element


if __name__ == '__main__':
    cli()



"""
Set individual bitflags needed for wofls.
"""
import numpy as np
import scipy.ndimage
import xarray

from wofs import terrain, constants, boilerplate
from wofs.constants import MASKED_CLOUD, MASKED_CLOUD_SHADOW, NO_DATA

PQA_SATURATION_BITS = sum(2 ** n for n in [0, 1, 2, 3, 4, 7])  # exclude thermal
PQA_CONTIGUITY_BITS = 0x01FF
PQA_CLOUD_BITS = 0x0C00
PQA_CLOUD_SHADOW_BITS = 0x3000
PQA_SEA_WATER_BIT = 0x0200


def dilate(array):
    """Dilation e.g. for cloud and cloud/terrain shadow"""
    # kernel = [[1] * 7] * 7 # blocky 3-pixel dilation
    y, x = np.ogrid[-3:4, -3:4]
    kernel = (x * x) + (y * y) <= 3.5 ** 2  # disk-like 3-pixel radial dilation
    return scipy.ndimage.binary_dilation(array, structure=kernel)


@boilerplate.simple_numpify
def pq_filter(pq):
    """
    Propagate flags from the pixel quality product.

    PQ specs: 16 bits.
      0-7 non-saturation of bands 1-5, 6.1, 6.2, 7. (Note bands 6 are thermal, irrelevent to standard WOfS.)
      8 contiguity (presumably including thermal bands)
      9 land (versus sea)
      10-11 no cloud (ACCA, Fmask)
      12-13 no cloud shadow (ACCA, Fmask)
      14 topographic shadow (not implemented)
      15 unspecified

     Over/under-saturation is flagged in the WOfS journal paper, but may not be previously implemented.

     Notes:
       - will output same flag to indicate noncontiguity, oversaturation and undersaturation.
       - disregarding PQ contiguity flag (see eo_filter instead) to exclude thermal bands.
       - permitting simultaneous flags (through addition syntax) since constants happen to be
         different powers of the same base.
       - dilates the cloud and cloud shadow. (Previous implementation eroded the negation.)
       - input must be numpy not xarray.DataArray (due to depreciated boolean fancy indexing behaviour)
    """
    ipq = ~pq  # bitwise-not, e.g. flag cloudiness rather than cloudfree

    masking = np.zeros(ipq.shape, dtype=np.uint8)
    masking[
        (ipq & (PQA_SATURATION_BITS | PQA_CONTIGUITY_BITS)).astype(np.bool)
    ] = constants.MASKED_NO_CONTIGUITY
    # masking[(ipq & PQA_SEA_WATER_BIT).astype(np.bool)] += constants.MASKED_SEA_WATER
    masking[dilate(ipq & PQA_CLOUD_BITS)] += constants.MASKED_CLOUD
    masking[dilate(ipq & PQA_CLOUD_SHADOW_BITS)] += constants.MASKED_CLOUD_SHADOW
    return masking


# C2_NODATA_BITS = 0x0001  # 0001 0th bit
C2_DILATED_BITS = 0x0002  # 0010 1st bit
C2_CLOUD_BITS = 0x0008  # 1000 3rd bit
C2_CLOUD_SHADOW_BITS = 0x0010  # 0001 0000 4th bit
# C2_CLEAR_BITS = 0x0040
C2_CIRRUS_BITS = 0x0004  # 0100 2nd bit


def c2_filter(pq):
    """
    Propagate flags from the pixel quality product.

    PQ specs: 16 bits.
      0 no data # dec 1
      1 dilated cloud # dec 2
      2 cirrus # dec 4
      3 cloud # dec 8
      4 cloud shadow # dec 16
      5 snow # dec 32
      6 clear # dec 64
      7 water # dec 128
      8-9 cloud confidence # dec 256/512
      10-11 cloud shadow confidence # dec 1024/2048
      12-13 snow_ice confidence # dec 4096/8192
      14-15 cirrus confidence # dec 16384/32768

     Notes:
       - permitting simultaneous flags (through addition syntax) since constants happen to be
         different powers of the same base.
       - dilates the cloud shadow. (cloud already dilated.)
       - input must be numpy not xarray.DataArray (due to depreciated boolean fancy indexing behaviour)
    """

    masking = np.zeros(pq.shape, dtype=np.uint8)
    masking[
        ((pq & C2_DILATED_BITS)).astype(np.bool)
        | ((pq & C2_CLOUD_BITS)).astype(np.bool)
        | ((pq & C2_CIRRUS_BITS)).astype(np.bool)
    ] += constants.MASKED_CLOUD
    masking[dilate(pq & C2_CLOUD_SHADOW_BITS)] += constants.MASKED_CLOUD_SHADOW
    return masking


def terrain_filter(dsm, nbar, no_data=-1000, ignore_dsm_no_data=False):
    """Terrain shadow masking, slope masking, solar incidence angle masking.

    Args:
        dsm: An XArray Dataset
        nbar: a Dataset that can be used to get a time
        no_data: NoDATA value from the DSM, defaults to -1000
        ignore_dsm_no_data: If True, don't flag nodata areas as shadow
    """

    shadows, slope, sia = terrain.shadows_and_slope(
        dsm, nbar.blue.time.values, no_data=no_data
    )

    # Alex Leith 2021: Assuming that the intention is that nodata
    # in the DSM means nodata in the WOfS. I'm making this
    # an option so we can include dsm no_data areas.
    if ignore_dsm_no_data:
        shadowy = dilate(shadows == terrain.SHADED)
    else:
        shadowy = dilate(shadows != terrain.LIT)

    low_sia = sia < constants.LOW_SOLAR_INCIDENCE_THRESHOLD_DEGREES

    steep = slope > constants.SLOPE_THRESHOLD_DEGREES

    result = (
        np.uint8(constants.MASKED_TERRAIN_SHADOW) * shadowy
        | np.uint8(constants.MASKED_HIGH_SLOPE) * steep
        | np.uint8(constants.MASKED_LOW_SOLAR_ANGLE) * low_sia
    )

    return xarray.DataArray(
        result, coords=[dsm.y, dsm.x]
    )  # note, assumes (y,x) axis ordering


def eo_filter(source):
    """
    Find where there is no data

    Input must be dataset, not array (since bands could have different nodata values).

    Contiguity can easily be tested either here or using PQ.
    """
    nodata_bools = source.map(lambda array: array == array.nodata).to_array(dim="band")

    nothingness = nodata_bools.all(dim="band")
    noncontiguous = nodata_bools.any(dim="band")

    return (
        np.uint8(constants.NO_DATA) * nothingness
        | np.uint8(constants.MASKED_NO_CONTIGUITY) * noncontiguous
    )


def fmask_filter(fmask):
    masking = np.zeros(fmask.shape, dtype=np.uint8)
    masking[fmask == 0] += NO_DATA
    masking[fmask == 2] += MASKED_CLOUD
    masking[fmask == 3] += MASKED_CLOUD_SHADOW

    return masking



import datacube
import geopandas
import rasterio
import rasterio.features
import matplotlib.pyplot as plt
import xarray
import numpy as np
import scipy
import scipy.ndimage
import glob
import sklearn.linear_model
import pickle

"""
# These weights are specified in the WOfS journal paper.
wofl_wet_freq = 0.1703
MrVBF = 0.1671
MODIS_OWL = 0.0336
slope = -0.2522
geofabric_foreshore = 4.2062
geofabric_pondage = -5.4692
geofabric_reservoir = 0.6574
geofabric_flat = 0.7700
geofabric_lake = 1.9992
geofabric_swamp = 1.3231
geofabric_watercourse = 1.9206
urban_areas = -4.9358

threshold = 0.05 # also by WOfS journal paper. Actually, maybe 0.01 (or 0.02)
"""


# --------------- Utilities ---------------


def numpy_to_xarray(array, geobox, name=None):
    """Utility to convert ndarray to DataArray, using a datacube.model.GeoBox"""
    coords = [xarray.IndexVariable(x, geobox.coords[x].values, attrs=dict(units=geobox.coords[x].units))
              for x in geobox.dims]
    return xarray.DataArray(array, coords=coords, attrs=dict(crs=geobox.crs), name=name)


def geopandas_to_xarray(table, geobox, name=None):
    """Rasterise (with reprojection)"""
    array = rasterio.features.rasterize(shapes=table.to_crs(geobox.crs._crs.ExportToProj4()).geometry,
                                        out_shape=(geobox.height, geobox.width),
                                        transform=geobox.affine)
    return numpy_to_xarray(array, geobox, name)


# Note on dimensions:
#   Rasterio has named height and width parameters,
#   referring to the first and second index of the raster array respectively.
#   ODC GeoBox names width and height (reverse order) in its constructor,
#   however height is associated with the "y" axis internally, which
#   still becomes the first of the two axes in the xarray.
def rasterfile_to_xarray(file, geobox=None, name=None, nodata=None):
    """Blit like"""
    with rasterio.open(file) as src:
        assert src.indexes == (1,)  # assume single band
        if geobox is None:
            from datacube.utils.geometry import GeoBox, CRS

            # TODO: fix this heinousness
            global crs
            global affine
            crs = src.crs
            affine = src.transform

            geobox = GeoBox(width=src.width,
                            height=src.height,
                            affine=src.affine,  # .transform is a list, .affine is an object
                            crs=CRS(src.crs.wkt))
            array = src.read(1)
        else:
            band = rasterio.band(src, 1)  # do not attempt to read entire extent into memory
            array = np.empty((geobox.height, geobox.width), dtype=band.dtype)
            rasterio.warp.reproject(source=band,
                                    destination=array,
                                    dst_crs=geobox.crs.crs_str,
                                    dst_transform=geobox.affine,
                                    dst_nodata=nodata)
    return numpy_to_xarray(array, geobox, name)


def write(filename, data, nodata=None):
    """ Output raster to filesystem """
    # TODO: eliminate globals (crs,affine) and magic constants (dimensions)
    import rasterio
    with rasterio.open(filename,
                       mode='w',
                       width=4000,
                       height=4000,
                       count=1,
                       dtype=data.dtype.name,
                       driver='GTIFF',
                       nodata=nodata,
                       tiled=True,
                       compress='LZW',  # balance IO volume and CPU speed
                       affine=affine,
                       crs=crs) as destination:
        destination.write(data.data, 1)


# ------------------ Ancilliaries ------------------
def urban(geobox):
    ucl_path = "/g/data/v10/wofs/ancillary/ucl/UCL_2011_AUST.shp"
    u = geopandas.read_file(ucl_path)  # load shapes table
    u = u[u['SOS_NAME11'] == 'Major Urban']  # filter out <100k
    u = u.to_crs(geobox.crs._crs.ExportToProj4())  # reproject
    array = rasterio.features.rasterize(shapes=u.geometry,
                                        out_shape=(geobox.height, geobox.width),
                                        transform=geobox.affine)
    return numpy_to_xarray(array, geobox, 'urban')


"""
Geofabric structure:

SrcFCName, SrcFType, AHGFFType

AHGFWaterbody
    Flats ['Swamp'] [27]
    Lakes ['Lake'] [26]
    Reservoirs ['Town Rural Storage', 'Flood Irrigation Storage'] [25]

AHGFHydroArea
    CanalAreas ['Canal Area'] [55]
    Flats ['Land Subject To Inundation', 'Saline Coastal Flat', 'Marine Swamp'] [56]
    ForeshoreFlats ['Foreshore Flat'] [59]
    PondageAreas ['Settling Pond', 'Aquaculture Area', 'Salt Evaporator'] [57]
    RapidAreas ['Rapid Area'] [58]
    WatercourseAreas ['Watercourse Area'] [54]
"""


def geofabric_parts(geobox):
    geofabric_path = "/g/data/v10/wofs/ancillary/geofabric/SH_Cartography_GDB/SH_Cartography.gdb"
    # fiona.listlayers(geofabric_path)
    for layer in ['AHGFHydroArea', 'AHGFWaterbody']:  # consider these two layers
        table = geopandas.read_file(geofabric_path, layer=layer)[['AHGFFType', 'SrcFCName', 'SrcFType', 'geometry']]
        for fc, df in table.groupby('SrcFCName'):
            if fc not in ['CanalAreas', 'RapidAreas']:  # exclude these two feature classes
                name = layer + fc
                yield geopandas_to_xarray(df, geobox, name=name)


# def geofabric(geobox):
#    geofabric_weights = {'AHGFHydroAreaFlats': geofabric_flat, # includes marine swamp
#                         'AHGFHydroAreaForeshoreFlats': geofabric_foreshore,
#                         'AHGFHydroAreaPondageAreas': geofabric_pondage,
#                         'AHGFHydroAreaWatercourseAreas': geofabric_watercourse,
#                         'AHGFWaterbodyFlats': geofabric_swamp, # type 27
#                         'AHGFWaterbodyLakes': geofabric_lake,
#                         'AHGFWaterbodyReservoirs': geofabric_reservoir}
#    return sum(array*geofabric_weights[array.name] for array in geofabric_parts(geobox))
datacube.config.LocalConfig.db_database = 'wofstest'
datacube.config.LocalConfig.db_hostname = 'agdcstaging-db.nci.org.au'
dc = datacube.Datacube()  # only used for elevation


def slope_degrees(geobox):
    pad = 5  # pixels of margin buffering
    padded = geobox[-pad:(geobox.height + pad), -pad:(geobox.width + pad)]

    # dem = dc.load(product='dsm1sv10', # ? 'srtm_dem1sv1_0'
    #              geopolygon=geobox[-pad:geobox.height+pad, -pad:geobox.width+pad].extent,
    #              output_crs=geobox.crs.crs_str, # force target gridspec, address weird bug
    #              resolution=geobox.resolution).isel(time=0)
    class likely:
        geobox = padded

    likely.extent = likely.geobox.extent
    likely.coords = likely.geobox.coords
    dem = dc.load(product='dsm1sv10',  # ? 'srtm_dem1sv1_0'
                  like=likely()).isel(time=0)

    # Sobel is prefered gradient method from DEM-grid literature.
    xgrad = scipy.ndimage.sobel(dem.elevation, axis=1) / abs(8 * dem.affine.a)  # i.e. dz/dx
    ygrad = scipy.ndimage.sobel(dem.elevation, axis=0) / abs(8 * dem.affine.e)
    # Here, assuming orthogonal grid. Probably shouldn't.

    # slope = numpy.degrees(numpy.arctan(numpy.hypot(xgrad, ygrad)))
    slope = np.degrees(np.arccos(1.0 / np.sqrt(xgrad ** 2 + ygrad ** 2 + 1.0)))
    # Tangential vectors have basis x+dz/dx z, y+dz/dy z.
    # Perpendicularity implies normal is colinear with z - dz/dx x - dz/dy y.
    # The slope cosine is given by the dot product of the normal with vertical
    # (i.e. by the vertical component of the normal, after magnitude normalisation).
    # Note, an alternative approach is to project the normal into the horizontal plane
    # (delaying magnitude normalisation until afterward),
    # and consider the rise in this direction of steepest ascent (seems like fewer operations).

    return numpy_to_xarray(slope[pad:-pad, pad:-pad], geobox, 'slope')  # strip padding


mrvbf_path = "/g/data/v10/wofs/ancillary/mrvbf/mrvbf_int.tif"
modis_path = "/g/data/v10/wofs/ancillary/modis/MOD09A1.aust.005.OWL.0.2.4.2001.2010.GE.20.tif"


# def synthesis(frequency_raster_path):
#
#    freq = rasterfile_to_xarray(frequency_raster_path)
#    geobox = freq.geobox
#
#    ucl = urban(geobox)
#    g = geofabric(geobox)
#    slant = slope_degrees(geobox)
#    vbf = rasterfile_to_xarray(mrvbf_path, geobox) # let nodata remain at 255
#    owl = rasterfile_to_xarray(modis_path, geobox)
#
#    ancillary = g + slope*slant + MODIS_OWL*owl + MrVBF*vbf + urban_areas*ucl
#
#    freq.data = np.nan_to_num(freq.data) # NaN -> 0
#    confidence = scipy.special.expit(ancillary + wofl_wet_freq*freq)
#
#    # filtered summary
#    filtered = freq*(confidence>threshold)
#
#    return confidence, filtered
def ancilliary_experts(geobox):
    ucl = urban(geobox)
    slant = slope_degrees(geobox)
    vbf = rasterfile_to_xarray(mrvbf_path, geobox, 'MrVBF')  # let nodata remain at 255
    owl = rasterfile_to_xarray(modis_path, geobox, 'MODIS-OWL')
    g = list(geofabric_parts(geobox))

    full_set = g + [ucl, slant, vbf, owl]

    # should sort this, say by name?
    # option to concat into a single xarray
    return full_set


def synthesis(frequency_raster_path=None, geobox=None, return_freq=False):
    """Return collection of all input datasets"""
    if frequency_raster_path is None:
        frequency_raster_path = '/g/data/v10/testing_ground/wofs_summary/frequency.vrt'
        assert geobox is not None

    raw_freq = rasterfile_to_xarray(frequency_raster_path, geobox=geobox, name='freq')

    if geobox is None:
        geobox = raw_freq.geobox

    everything = ancilliary_experts(geobox) + [raw_freq]

    result = xarray.concat(everything, dim='variable')
    return result if not return_freq else (result, raw_freq)


def training_data(max_tiles=None):
    """
    This gets the synthesis of data over the training areas,
    and unmasked subset (e.g. clear count must also be nonzero)
    """

    def generate_parts():
        for area in glob.glob('/g/data/u46/wofs/confidence/Training/*.tif')[:max_tiles]:
            training = rasterfile_to_xarray(area, name='training')

            clear_path = '/g/data/v10/testing_ground/wofs_summary/clear.vrt'
            clear = rasterfile_to_xarray(clear_path, geobox=training.geobox, name='clear_counts')

            mask = (clear.data > 0) & (training.data != 0)
            # del clear

            data = synthesis(geobox=training.geobox)

            Y_subset = (training.data[mask] == 2)
            T_subset = data.data[np.broadcast_to(mask, data.shape)].reshape((len(data), len(Y_subset)))

            yield Y_subset, T_subset

    Ys, Ts = map(np.hstack, zip(*generate_parts()))

    return Ys, Ts


model_path = '/g/data/v10/testing_ground/wofs_summary/logistic_model.pkl'


def train():
    print('Loading..')

    Ys, Ts = training_data()  # 8 works, 10 the kernel dies.

    model = sklearn.linear_model.LogisticRegression(tol=1e-7, class_weight='balanced')

    print('Fitting..')

    model.fit(Ts.T, Ys)

    print('Coefficients:')
    print(model.coef_)
    print('Intercept:')
    print(model.intercept_)

    with open(model_path, 'wb') as f:
        pickle.dump(model, f)

    return model


def process(filename, model, threshold=0.5):
    assert "frequency" in filename
    out1 = filename.replace("frequency", "confidence")
    out2 = filename.replace("frequency", "filtered")
    import os.path
    assert not os.path.isfile(out1)
    assert not os.path.isfile(out2)
    factors, freq = synthesis(filename, return_freq=True)
    X = factors.data.reshape((len(factors), -1)).T
    X = np.nan_to_num(X)  # NaN -> zero
    P = model.predict_proba(X)
    conf = P[:, 1].reshape(freq.shape)
    conf = numpy_to_xarray(conf, freq.geobox)

    write(out1, conf.astype(np.float32))

    freq.data[conf.data <= threshold] = np.nan
    filt = freq

    write(out2, filt.astype(np.float32), nodata=np.nan)


if __name__ == '__main__':
    import sys

    if len(sys.argv) < 2:
        print("Usage: python confidence.py *frequency*.tif")
        print("Example: /usr/bin/time find /g/data/v10/testing_ground/wofs_summary/ -maxdepth 1 "
              "-name */*frequency.tif | xargs -n 10 -P8 python confidence.py")
        raise SystemExit
    elif len(sys.argv) == 2 and sys.argv[1] == '--retrain':
        model = train()
        print('Done.')
    else:
        with open(model_path, 'rb') as f:
            model = pickle.load(f)
        for filename in sys.argv[1:]:
            process(filename, model)



import logging
from typing import Dict

import xarray as xr
from datacube.testutils.io import dc_read
from datacube.virtual import Transformation, Measurement
from xarray import Dataset

from wofs.wofls import woffles_ard, woffles_usgs_c2

WOFS_OUTPUT = [{
    'name': 'water',
    'dtype': 'uint8',
    'nodata': 1,
    'units': '1'
}, ]
_LOG = logging.getLogger(__file__)


def scale_usgs_collection2(data):
    """These are taken from the Fractional Cover scaling values"""
    attrs = data.attrs
    data =  data.apply(scale_and_clip_dataarray, keep_attrs=False,
                       scale_factor=0.275, add_offset=-2000,
                       clip_range=None, valid_range=(0, 10000))
    data.attrs = attrs
    return data


def scale_and_clip_dataarray(dataarray: xr.DataArray, *, scale_factor=1, add_offset=0, clip_range=None,
                             valid_range=None, new_nodata=-999, new_dtype='int16'):
    orig_attrs = dataarray.attrs
    nodata = dataarray.attrs['nodata']

    mask = dataarray.data == nodata

    # add another mask here for if data > 10000 then also make that nodata
    dataarray = dataarray * scale_factor + add_offset

    if clip_range is not None:
        dataarray = dataarray.clip(*clip_range)

    dataarray = dataarray.astype(new_dtype)

    dataarray.data[mask] = new_nodata
    if valid_range is not None:
        valid_min, valid_max = valid_range
        dataarray = dataarray.where(dataarray >= valid_min, new_nodata)
        dataarray = dataarray.where(dataarray <= valid_max, new_nodata)
    dataarray.attrs = orig_attrs
    dataarray.attrs['nodata'] = new_nodata

    return dataarray


def _to_xrds_coords(geobox):
    return {dim: coord.values for dim, coord in geobox.coordinates.items()}


class WOfSClassifier(Transformation):
    """ Applies the wofs algorithm to surface reflectance data.
    Requires bands named
    bands = ['nbart_blue', 'nbart_green', 'nbart_red', 'nbart_nir', 'nbart_swir_1', 'nbart_swir_2', 'fmask']

    Terrain buffer is specified in CRS Units (typically meters)

    Options include:
        dsm_path: a URI to a DSM, either S3:// or HTTPS:// work
        c2_scaling: handle the USGS's new scaling values, rescaling to the old way
        terrain_buffer:
    """

    def __init__(self, dsm_path=None, c2_scaling=False, terrain_buffer=0, dsm_no_data=-1000, ignore_dsm_no_data=False):
        self.dsm_path = dsm_path
        self.dsm_no_data = dsm_no_data
        self.c2_scaling = c2_scaling
        self.terrain_buffer = terrain_buffer
        self.ignore_dsm_no_data = ignore_dsm_no_data
        self.output_measurements = {m['name']: Measurement(**m) for m in WOFS_OUTPUT}
        if dsm_path is None:
            _LOG.warning('WARNING: Path or URL to a DSM is not set. Terrain shadow mask will not be calculated.')

    def measurements(self, input_measurements) -> Dict[str, Measurement]:
        return self.output_measurements

    def compute(self, data) -> Dataset:
        _LOG.info(data.geobox)
        _LOG.info(repr(data.geobox))

        if self.c2_scaling:
            # The C2 data need to be scaled
            orig_attrs = data.attrs
            spectral_data = data[
                ['nbart_blue', 'nbart_green', 'nbart_red', 'nbart_nir', 'nbart_swir_1', 'nbart_swir_2']
            ]
            mask_data = data[['fmask']]
            data = xr.merge([scale_usgs_collection2(spectral_data), mask_data])
            data.attrs = orig_attrs

        if self.dsm_path is not None:
            dsm = self._load_dsm(data.geobox.buffered(self.terrain_buffer, self.terrain_buffer))
        else:
            dsm = None

        wofs = []
        for time_idx in range(len(data.time)):
            if self.c2_scaling:
                # C2 wofls
                wofs.append(
                    woffles_usgs_c2(
                        data.isel(time=time_idx),
                        dsm,
                        dsm_no_data=self.dsm_no_data,
                        ignore_dsm_no_data=self.ignore_dsm_no_data
                    ).to_dataset(name='water')
                )
            else:
                wofs.append(
                    woffles_ard(
                        data.isel(time=time_idx),
                        dsm,
                        dsm_no_data=self.dsm_no_data,
                        ignore_dsm_no_data=self.ignore_dsm_no_data
                    ).to_dataset(name='water')
                )

        wofs = xr.concat(wofs, dim='time')
        wofs.attrs['crs'] = data.attrs['crs']
        return wofs

    def _load_dsm(self, gbox):
        # Data variable needs to be named elevation
        dsm = dc_read(self.dsm_path, gbox=gbox, resampling="bilinear")
        return xr.Dataset(
            data_vars={'elevation': (('y', 'x'), dsm)},
            coords=_to_xrds_coords(gbox),
            attrs={'crs': gbox.crs}
        )



# Josh Sixsmith, refactored by BL.
import numpy
import logging
import gc

try:
    import dask.array
    dask_array_type = (dask.array.Array,)
except ImportError:  # pragma: no cover
    dask_array_type = ()

from wofs import boilerplate


@boilerplate.simple_numpify
def classify(images, float64=False):
    if isinstance(images, dask_array_type):
        # Apply the classify function on each block in the x and y dimensions
        # Remove chunks and reduce along the 'band' dimension (axis 0)
        return dask.array.map_blocks(_classify, images.rechunk({0: -1}), drop_axis=0, dtype='uint8')
    return _classify(images, float64)


# pylint: disable=too-many-locals,too-many-statements
def _classify(images, float64=False):
    """
    Produce a water classification image from the supplied images (6 bands of an NBAR, multiband Landsat image)
    This method evaluates N.Mueller's decision tree as follows:

                    -----------------------------N1---------------------------------
                    |                                                              |
                    |                                                              |
                 ---N2-----                                           -------------N21---------------------
                 |        |                                           |                                   |
                 |        |                                           |                                   |
       ----------N4----   N3                                    ------N22---                           ---N35-------
       |              |                                         |          |                           |           |
       |              |                                         |          |                           |           |
    ---N5---       ---N8--------------                       ---N24----    N23                      ---N37------   N36
    |      |       |                 |                       |        |                             |          |
    |      |       |                 |                       |        |                             |          |
    N6     N7   ---N12------------   N9             ---------N26---   N25                        ---N39-----   N38
                |                |                  |             |                              |         |
                |                |                  |             |                              |         |
             ---N16---        ---N13---             N27   --------N28---                   ------N41---    N40
             |       |        |       |                   |            |                   |          |
             |       |        |       |                   |            |                   |          |
             N17  ---N18---   N14     N15              ---N29---    ---N30---           ---N43---     N42
                  |       |                            |       |    |       |           |       |
                  |       |                            |       |    |       |           |       |
                  N19     N20                          N31     N32  N33     N34         N44     N45

    :param images:
        A 3D numpy array ordered in (bands,rows,columns), containing the spectral data.
        It is assumed that the spectral bands follow Landsat 5 & 7, Band 1, Band 2, Band 3, Band 4, Band 5, Band 7.

    :param float64:
        Boolean keyword. If set to True then the data will be converted to type float64 if not already float64.
        Default is False.

    :return:
        A 2D numpy array of type UInt8.  Values will be 0 for No Water, 1 for Unclassified and 128 for water.

    :notes:
        The input array will be converted to type float32 if not already float32.
        If images is of type float64, then images datatype will be left as is.

    :transcription:
        Transcribed from a Tree diagram output by CART www.salford-systems.com
        Josh Sixsmith; joshua.sixsmith@ga.gov.au

    """

    logger = logging.getLogger("WaterClasserfier")  # !? typo..
    logger.debug("Started")

    def band_ratio(a, b):
        """
        Calculates a normalised ratio index.
        """
        c = (a - b) / (a + b)
        return c

    dims = images.shape
    if len(dims) == 3:
        bands = dims[0]
        rows = dims[1]
        cols = dims[2]
    else:
        rows = dims[0]
        cols = dims[1]

    dtype = images.dtype

    # Check whether to enforce float64 calcs, unless the datatype is already float64
    # Otherwise force float32
    if float64:
        if dtype != 'float64':
            images = images.astype('float64')
    else:
        if dtype == 'float64':
            # Do nothing, leave as float64
            images = images
        elif dtype != 'float32':
            images = images.astype('float32')

    classified = numpy.ones((rows, cols), dtype='uint8')

    ndi_52 = band_ratio(images[4], images[1])
    ndi_43 = band_ratio(images[3], images[2])
    ndi_72 = band_ratio(images[5], images[1])

    b1 = images[0]
    b2 = images[1]
    b3 = images[2]
    b4 = images[3]
    b5 = images[4]
    b7 = images[5]

    # Lets start going down the trees left branch, finishing nodes as needed
    # Lots of result arrays eg r1, r2 etc of type bool are created
    # These could be recycled to save memory, but at the moment they serve to show the tree structure
    # Temporary arrays of type bool (_tmp, _tmp2) are used to combine the boolean decisions
    r1 = ndi_52 <= -0.01

    r2 = b1 <= 2083.5
    classified[r1 & ~r2] = 0  # Node 3

    r3 = b7 <= 323.5
    _tmp = r1 & r2
    _tmp2 = _tmp & r3
    _tmp &= ~r3

    r4 = ndi_43 <= 0.61
    classified[_tmp2 & r4] = 128  # Node 6
    classified[_tmp2 & ~r4] = 0  # Node 7

    r5 = b1 <= 1400.5
    _tmp2 = _tmp & ~r5
    r6 = ndi_43 <= -0.01
    classified[_tmp2 & r6] = 128  # Node 10
    classified[_tmp2 & ~r6] = 0  # Node 11

    _tmp &= r5

    r7 = ndi_72 <= -0.23
    _tmp2 = _tmp & ~r7
    r8 = b1 <= 379
    classified[_tmp2 & r8] = 128  # Node 14
    classified[_tmp2 & ~r8] = 0  # Node 15

    _tmp &= r7

    r9 = ndi_43 <= 0.22
    classified[_tmp & r9] = 128  # Node 17

    _tmp &= ~r9

    r10 = b1 <= 473
    classified[_tmp & r10] = 128  # Node 19
    classified[_tmp & ~r10] = 0  # Node 20

    # Left branch is completed; cleanup
    logger.debug("B4 cleanup 1")
    del r2, r3, r4, r5, r6, r7, r8, r9, r10
    gc.collect()
    logger.debug("cleanup 1 done")

    # Right branch of the tree
    r1 = ~r1

    r11 = ndi_52 <= 0.23
    _tmp = r1 & r11

    r12 = b1 <= 334.5
    _tmp2 = _tmp & ~r12
    classified[_tmp2] = 0  # Node 23

    _tmp &= r12

    r13 = ndi_43 <= 0.54
    _tmp2 = _tmp & ~r13
    classified[_tmp2] = 0  # Node 25

    _tmp &= r13

    r14 = ndi_52 <= 0.12
    _tmp2 = _tmp & r14
    classified[_tmp2] = 128  # Node 27

    _tmp &= ~r14

    r15 = b3 <= 364.5
    _tmp2 = _tmp & r15

    r16 = b1 <= 129.5
    classified[_tmp2 & r16] = 128  # Node 31
    classified[_tmp2 & ~r16] = 0  # Node 32

    _tmp &= ~r15

    r17 = b1 <= 300.5
    _tmp2 = _tmp & ~r17
    _tmp &= r17
    classified[_tmp] = 128  # Node 33
    classified[_tmp2] = 0  # Node 34

    _tmp = r1 & ~r11

    r18 = ndi_52 <= 0.34
    classified[_tmp & ~r18] = 0  # Node 36
    _tmp &= r18

    r19 = b1 <= 249.5
    classified[_tmp & ~r19] = 0  # Node 38
    _tmp &= r19

    r20 = ndi_43 <= 0.45
    classified[_tmp & ~r20] = 0  # Node 40
    _tmp &= r20

    r21 = b3 <= 364.5
    classified[_tmp & ~r21] = 0  # Node 42
    _tmp &= r21

    r22 = b1 <= 129.5
    classified[_tmp & r22] = 128  # Node 44
    classified[_tmp & ~r22] = 0  # Node 45

    logger.debug("completed")

    return classified



import xarray


def simple_numpify(f):
    """Transform a numpy operation to an xarray DataArray operation
    Assumes only (y,x) arrays."""
    def wrapped(xr):
        return xarray.DataArray(f(xr.data), coords=[xr.y, xr.x])
        # return xarray.DataArray(f(xr.data), coords=[x[c] for c in list(x.dims) if c in {'y','x'}])
    wrapped.__name__ = f.__name__
    return wrapped



import math

import ephem
import numpy
import xarray
from datacube.utils.geometry import CRS, line
from pandas import to_datetime
from scipy import ndimage

UNKNOWN = -1
LIT = 255
SHADED = 0


def _shade_row(shade_mask, elev_m, sun_alt_deg, pixel_scale_m, no_data, fuzz=0.0):
    """
    shade the supplied row of the elevation model
    """

    # threshold is TAN of sun's altitude
    tan_sun_alt = math.tan(sun_alt_deg)

    # pure terrain angle shadow
    shade_mask[0] = LIT
    shade_mask[1:] = numpy.where((elev_m[:-1] - elev_m[1:]) / pixel_scale_m < tan_sun_alt, LIT, SHADED)

    # project shadows from tips (light->shadow transition)
    switch = numpy.where(shade_mask[:-1] != shade_mask[1:])
    for i in switch[0]:  # note: could use flatnonzero instead of where; or else switch,=; to avoid [0]. --BL
        if shade_mask[i] == LIT:
            # TODO: horizontal fuzz?
            shadow_level = (elev_m[i] + fuzz) - numpy.arange(shade_mask.size - i) * (tan_sun_alt * pixel_scale_m)
            shade_mask[i:][shadow_level > elev_m[i:]] = SHADED

    shade_mask[elev_m == no_data] = UNKNOWN

    return shade_mask


def vector_to_crs(point, vector, original_crs, destination_crs):
    """
    Transform a vector (in the tangent space of a particular point) to a new CRS

    Expects point and vector to each be a 2-tuple in the original CRS.
    Returns a pair of 2-tuples (transformed point and vector).
    Order of coordinates is specified by the CRS (or the OGR library).
    """
    # pylint: disable=zip-builtin-not-iterating
    # theoretically should use infinitesimal displacement
    # i.e. jacobian of the transformation
    # but here just use a finite displatement (for convenience of implementation)
    original_line = line([point, tuple(map(sum, zip(point, vector)))], crs=original_crs)
    transformed_line = original_line.to_crs(destination_crs)

    transformed_point, _ = transformed_line.points

    # take difference (i.e. remove origin offset)
    transformed_vector = tuple(map(lambda x: x[1] - x[0], zip(*transformed_line.points)))
    return transformed_point, transformed_vector


def solar_vector(point, time, crs):
    (lon, lat), (dlon, dlat) = vector_to_crs(point, (0, 100),
                                             original_crs=CRS(crs),
                                             destination_crs=CRS('EPSG:4326'))

    # azimuth north to east of the vertical direction of the crs
    vert_az = math.atan2(dlon * math.cos(math.radians(lat)), dlat)

    observer = ephem.Observer()
    # pylint: disable=assigning-non-slot
    observer.lat = math.radians(lat)
    observer.lon = math.radians(lon)
    observer.date = time
    sun = ephem.Sun(observer)

    sun_az = sun.az - vert_az
    x = math.sin(sun_az) * math.cos(sun.alt)
    y = -math.cos(sun_az) * math.cos(sun.alt)
    z = math.sin(sun.alt)

    return x, y, z, sun_az, sun.alt


# pylint: disable=too-many-locals
def shadows_and_slope(tile, time, no_data=-1000):
    """
    Terrain shadow masking (Greg's implementation) and slope masking.

    Input: Digital Surface Model xarray DataSet (need metadata e.g. resolution, CRS)

    Uses Sobel filter to estimate the slope gradients (assuming raster is non-rotated wrt. crs) and magnitude.
    Ignores curvature of earth (picking middle of tile for solar elevation and azimuth) calculating surface incidence.
    Reprojects (rotates/resamples) DSM to align rows with shadows (at 25m resolution,
    and assuming the input projection is Mercator-like i.e. preserves bearings).
    For each row, finds each threshold pixel (where the slope just turns away from the sun) and raytraces
    (i.e. using a ramp, masks the other pixels shaded by the pillar of that pixel).
    Reprojects shadow mask (and undoes border enlargement associated with the rotation).

    TODO (BL) -- profile, and explore numpy.minimum.accumulate (make-monotonic) style alternative
                 and maybe fewer resamplings (or come up with something better still).
    """

    y_size, x_size = tile.elevation.shape

    # row spacing
    pixel_scale_m = abs(tile.affine.e)

    # gradient and slope
    xgrad = ndimage.sobel(tile.elevation, axis=1) / abs(8 * tile.affine.a)
    ygrad = ndimage.sobel(tile.elevation, axis=0) / abs(8 * tile.affine.e)

    # length of the terrain normal vector
    norm_len = numpy.sqrt((xgrad * xgrad) + (ygrad * ygrad) + 1.0)
    slope = numpy.degrees(numpy.arccos(1.0 / norm_len))

    x, y = tile.dims.keys()
    tile_center = (tile[x].values[x_size // 2], tile[y].values[y_size // 2])
    solar_vec = solar_vector(tile_center, to_datetime(time), tile.crs)
    sia = (solar_vec[2] - (xgrad * solar_vec[0]) - (ygrad * solar_vec[1])) / norm_len
    sia = 90 - numpy.degrees(numpy.arccos(sia))

    rot_degrees = 90.0 + math.degrees(solar_vec[3])

    buff_elv_array = numpy.pad(tile.elevation.values, 4, mode='edge')
    rotated_elv_array = ndimage.interpolation.rotate(buff_elv_array,
                                                     rot_degrees,
                                                     reshape=True,
                                                     output=numpy.float32,
                                                     cval=no_data,
                                                     prefilter=False)

    # create the shadow mask by ray-tracying along each row
    shadows = numpy.zeros_like(rotated_elv_array)
    for row in range(0, rotated_elv_array.shape[0]):
        _shade_row(shadows[row], rotated_elv_array[row], solar_vec[4], pixel_scale_m, no_data, fuzz=10.0)

    del rotated_elv_array
    del buff_elv_array

    shadows = ndimage.interpolation.rotate(shadows, -rot_degrees, reshape=False, output=numpy.float32, cval=no_data,
                                           prefilter=False)

    dr = (shadows.shape[0] - y_size) // 2
    dc = (shadows.shape[1] - x_size) // 2

    shadows = shadows[dr:dr + y_size, dc:dc + x_size]
    shadows = xarray.DataArray(shadows.reshape(tile.elevation.shape), coords=tile.elevation.coords)

    return shadows, slope, sia



from pathlib import Path

import xarray as xr
import yaml
from datacube.virtual import construct

from wofs.virtualproduct import WOfSClassifier


def test_virtualproduct():
    # Load sample surface reflectance data
    sr_data = xr.open_dataset(Path(__file__).parent / 'sample_c3_sr.nc', mask_and_scale=False)
    # and munge to make look more like data loaded by ODC
    sr_data = sr_data.rename({'oa_fmask': 'fmask'})
    sr_data.attrs['crs'] = 'EPSG:32754'
    del sr_data.coords['band']

    for dv in sr_data.data_vars.values():
        dv.attrs['nodata'] = dv.attrs['nodatavals']

    transform = WOfSClassifier()
    wofl = transform.compute(sr_data)

    sample = xr.open_dataset(Path(__file__).parent / 'sample_wofl.nc', mask_and_scale=False)

    assert sample.equals(wofl)


def foo():
    # measurements: [green, red, nir, swir1, swir2]
    virtual_product_defn = yaml.safe_load('''
    transform: wofs.virtualproduct.Wofs
    input:
        product: ls8_ard
        measurements: [nbart_blue, nbart_green, nbart_red, nbart_nir, nbart_swir_1, nbart_swir_2, fmask]
    ''')
    virtual_product = construct(**virtual_product_defn)

    # [odc_conf_test] -
    # db_hostname: agdcdev-db.nci.org.au
    # db_port: 6432
    # db_database: odc_conf_test

    # [ard_interop] - collection upgrade DB
    # db_hostname: agdcstaging-db.nci.org.au
    # db_port:     6432
    # db_database: ard_interop

    dc = datacube.Datacube(env="odc_conf_test")

    vdbag = virtual_product.query(dc=dc, id='be43b7ce-c421-4c16-826d-a508f3e3d984')

    box = virtual_product.group(vdbag, output_crs='EPSG:28355', resolution=(-25, 25))

    virtual_product.output_measurements(vdbag.product_definitions)

    data = virtual_product.fetch(box, dask_chunks=dict(x=1000, y=1000))

    print(data)

    # crash!
    # done = data.compute()



import dask.array as da
import numpy as np
import pytest
import xarray as xr
from affine import Affine

from datacube.utils.geometry import GeoBox, CRS
from wofs.virtualproduct import WOfSClassifier


def test_nodata_bit_setting(sample_data):
    """

    If no-data bit (bit 1) is set, all other bits should be 0. -- Recommendation from Norman Mueller.
    """

    classifier = WOfSClassifier()

    wofl = classifier.compute(sample_data)
    wofl = wofl.compute()
    wofl = wofl.water.data.reshape(-1)

    values_with_nodata_bit_set = wofl[np.bitwise_and(wofl, 1) == 1]
    assert wofl.dtype == np.dtype('uint8')
    assert (values_with_nodata_bit_set == 1).all()


@pytest.fixture(params=[np, da], ids=["numpy", "dask.array"])
def sample_data(request):
    """Load sample Surface reflectance and QA data"""
    required_bands = ['nbart_blue', 'nbart_green', 'nbart_red', 'nbart_nir', 'nbart_swir_1', 'nbart_swir_2', 'fmask']

    test_data = xr.Dataset(
        {
            name: (['time', 'y', 'x'],
                   request.param.arange(0, 2 ** 16, dtype='uint16').reshape((1, 256, 256)),
                   {'nodata': 0})
            for name in required_bands
        }, attrs={
            'geobox': GeoBox(256, 256, Affine(0.01, 0.0, 139.95,
                                              0.0, -0.01, -49.05), CRS('EPSG:3577')),
            'crs': CRS('EPSG:3577')
        }, coords={
            'time': np.array(['2000-01-01'], dtype='datetime64')
        })
    test_data.fmask.attrs['flags_definition'] = {
        'nodata': {'bits': 0, 'values': {0: False, 1: True}},
        'valid_aerosol': {'bits': 1, 'values': {0: 'not_valid', 1: 'valid'}},
        'water': {'bits': 2, 'values': {0: 'not_water', 1: 'water'}},
        'cloud_or_cirrus': {'bits': 3, 'values': {0: 'not_cloud_or_cirrus',
                                                  1: 'cloud_or_cirrus'}},
        'cloud_shadow': {'bits': 4,
                         'values': {0: 'not_cloud_shadow', 1: 'cloud_shadow'}},
        'interpolated_aerosol': {'bits': 5,
                                 'values': {0: 'not_aerosol_interpolated',
                                            1: 'aerosol_interpolated'}},
        'aerosol_level': {'bits': [6, 7],
                          'values': {0: 'climatology', 1: 'low', 2: 'medium',
                                     3: 'high'}}}
    return test_data



from datacube import Datacube

from wofs.wofls import woffles
import pytest

WATER = 136  # note, this is wrong.
NOT_WATER = 8


@pytest.mark.skip("Requires running ODC Database + Data")
@pytest.mark.parametrize("query,expected", [
    # Australian Parliament House, generally not water
    (dict(lat=(-35.308, -35.309), lon=(149.124, 149.125), time=('2016-01-10', '2016-01-14')), NOT_WATER),
    # Middle of Lake Burley Griffith. Water
    (dict(lat=(-35.295, -35.296), lon=(149.137, 149.138), time=('2016-01-10', '2016-01-14')), WATER),
])
def test_woffles(query, expected):
    dc = Datacube(app='test_wofls')

    bands = ['blue', 'green', 'red', 'nir', 'swir1', 'swir2']  # inputs needed from EO data)
    source = dc.load(product='ls8_nbar_albers', measurements=bands, **query)
    pq = dc.load(product='ls8_pq_albers', like=source)
    dsm = dc.load(product='dsm1sv10', like=source, time=('1900-01-01', '2100-01-01'), resampling='cubic')

    wofls_output = woffles(*(x.isel(time=0) for x in [source, pq, dsm]))

    assert (wofls_output == expected).all()



import os

from setuptools import find_packages, setup

config_files = ["config/" + name for name in os.listdir("config")]
tests_require = [
    "pytest",
    "pytest-cov",
    "mock",
    "pylint",
    "hypothesis",
    "compliance-checker",
    "black",
]
extras_require = {
    "test": tests_require,
}

setup(
    name="wofs",
    setup_requires=["setuptools_scm"],
    use_scm_version=True,
    description="Water Observations from Space - Digital Earth Australia",
    long_description=open("README.rst", "r").read(),
    license="Apache License 2.0",
    url="https://github.com/GeoscienceAustralia/wofs",
    author="Geoscience Australia",
    maintainer="Geoscience Australia",
    maintainer_email="",
    packages=find_packages(),
    data_files=[("wofs/config", config_files)],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: GIS",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
    install_requires=["datacube", "scipy", "ephem", "xarray>=0.14.1"],
    tests_require=tests_require,
    extras_require=extras_require,
    entry_points={
        "console_scripts": [
            "datacube-wofs = wofs.wofs_app:cli",
        ]
    },
)



"""
Test some of the terrain masking functions
"""

import pytest
import hypothesis
from hypothesis import given
from hypothesis import strategies as st

from datacube.utils.geometry import CRS
from wofs.terrain import vector_to_crs

# Use slightly less than the projected boundary from
# https://spatialreference.org/ref/epsg/gda94-australian-albers/
points_3577 = st.tuples(st.integers(min_value=-5000000, max_value=-3000000),
                        st.integers(min_value=-1500000, max_value=2500000))

vectors = st.tuples(st.integers(min_value=-100, max_value=100),
                    st.integers(min_value=-100, max_value=100))

@hypothesis.settings(deadline=500)
@given(points_3577, vectors)
def test_vector_to_crs(orig_point, orig_vect):
    """
    Given a random point (in EPSG:3577) and vector, convert to EPSG:4326
    and back

    Ensure you get your starting values back (approximately)
    """
    inter_point, inter_vect = vector_to_crs(orig_point, orig_vect,
                                            original_crs=CRS('EPSG:3577'),
                                            destination_crs=CRS('EPSG:4326'))

    new_point, new_vect = vector_to_crs(inter_point, inter_vect,
                                        original_crs=CRS('EPSG:4326'),
                                        destination_crs=CRS('EPSG:3577'))

    assert orig_point == pytest.approx(new_point, abs=1e-6)
    assert orig_vect == pytest.approx(new_vect, abs=1e-6)



import sys
from pathlib import Path

import xarray as xr
import yaml


def create_sample_file():
    in_dir = Path(sys.argv[1])
    out_file = sys.argv[2]

    fname = list(in_dir.glob('*.odc-metadata.yaml'))[0]

    with open(fname) as fin:
        doc = yaml.safe_load(fin)

    to_load = [(name, in_dir / details['path'])
               for name, details in doc['measurements'].items()
               if name.startswith('nbart') or 'fmask' in name]
    to_load = sorted(to_load, key=lambda tup: tup[1])

    def load_band(fname):
        data = xr.open_rasterio(fname)
        return data.squeeze().sel(x=slice(None, None, 200), y=slice(None, None, 200))

    data_vars = {name: load_band(fname) for name, fname in to_load}

    ds = xr.Dataset(data_vars)

    for dv in ds.data_vars.values():
        dv.attrs['nodata'] = dv.attrs['nodatavals']

    ds = ds.expand_dims({'time': [doc['properties']['datetime'].replace(tzinfo=None)]})

    ds.to_netcdf(out_file)


if __name__ == '__main__':
    create_sample_file()



"""
    Wet/Clear observation counts.

    Recommended usage: see job.sh

    Performance:
     - YAML parsing is slow. Could test obtaining JSON from DB index (and forego globbing)
     - Array operations could be optimised with numexpr (even with broadcasting for dual output),
       but requires future (e.g. v3) numexpr release to support relevant bitwise operations.

    TODO: record which datasets are utilised, and facilitate incrementally augmenting previous
    outputs as new datasets become available.

"""

import numpy as np
import functools


class reader:
    """ Read one water-extents tile/dataset """
    cell = None

    def __init__(self, file):
        self.path = file
        if self.cell is None:
            self.configure_cell(file)  # Assume all files will cover same spatial cell

    @classmethod
    def configure_cell(cls, example):
        import rasterio
        with rasterio.open('NetCDF:' + example + ':water') as f:
            cls.cell = dict(affine=f.profile['affine'], crs=f.profile['crs'])

    @property
    def netcdf(self):
        import netCDF4
        return netCDF4.Dataset(self.path)

    @property
    def metadata_doc(self):
        return self.netcdf['dataset'][0].tostring().decode('unicode_escape')

    @property
    @functools.lru_cache()
    def metadata(self):
        # This is slow (almost one second; quarter-second with C version)
        # TODO: consider querying database rather than file.
        # i.e. get JSON, which is faster to parse.
        import yaml
        return yaml.load(self.metadata_doc, Loader=yaml.CLoader)

    @property
    def timestamp(self):
        return self.metadata['extent']['center_dt']

    @property
    def date(self):
        import datetime
        import dateutil.parser
        # rough conversion to local date
        return (dateutil.parser.parse(self.timestamp) + datetime.timedelta(hours=10)).date()

    @property
    def gqa(self):
        import math
        offset = self.metadata['lineage']['source_datasets']['0']['lineage']['source_datasets']['0']['lineage'][
            'source_datasets']['level1']['gqa']['residual']['iterative_mean']
        return math.hypot(float(offset['x']), float(offset['y']))  # Or lookup pre-computed "xy" attribute
        # return self.metadata \
        #    ['lineage']['source_datasets']['0'] \
        #    ['lineage']['source_datasets']['0'] \
        #    ['lineage']['source_datasets']['level1'] \
        #    ['gqa']['residual']['cep90']

    @property
    def water(self):
        return np.squeeze(self.netcdf['water'])


class fuser:
    """ Combine one or more datasets into a single-rasterisable package """

    def __init__(self, tiles):
        self.tiles = tiles
        # print(len(tiles),end='')

    @property
    def water(self):
        output = self.tiles[0].water
        for tile in self.tiles[1:]:
            # TODO: could numexpr
            subsequent = tile.water
            empty = (output & 1).astype(np.bool)
            both = ~empty & ~((subsequent & 1).astype(np.bool))
            output[empty] = subsequent[empty]
            output[both] |= subsequent[both]
        return output


def do_work(observations):
    """ Aggregate data while reading one whole file at a time into memory """
    import numpy as np

    wet_accumulator = np.zeros((4000, 4000), np.uint16)
    dry_accumulator = np.zeros((4000, 4000), np.uint16)

    land_or_sea = ~np.uint8(4)  # to mask out the marine flag
    # wet_versus_dry = np.array([128,0])[None,None,:]

    for f in observations:
        bitfield = f.water & land_or_sea

        wet_accumulator += bitfield == 128
        dry_accumulator += bitfield == 0
        # print('.', end='')
    # print('')
    return wet_accumulator, dry_accumulator


def summarise_result(observations, prefix=''):
    """ Perform counts and from those produce outputs """
    wet, dry = do_work(observations)

    clear_observation_count = wet + dry

    import numpy as np
    with np.errstate(invalid='ignore'):  # denominator may be zero
        frequency = wet / clear_observation_count

    write(prefix + 'clear.tif', clear_observation_count)
    write(prefix + 'wet.tif', wet)
    write(prefix + 'frequency.tif', frequency.astype(np.float32), nodata=np.nan)

    """
    import matplotlib.pyplot as plt
    fig, (ax1,ax2) = plt.subplots(1,2)
    ax1.imshow(clear_observation_count[::10,::10])
    ax2.imshow(frequency[::10,::10])
    #fig.show()"""


def get_filenames(*directories):
    from glob import glob
    return [x for d in directories for x in glob(d + '/*.nc')]


def get_observations(files):
    """ Group datasets into observations, and perform GQA filtering """
    import pandas

    rr = map(reader, files)

    p = [(r, r.date) for r in rr if r.gqa < 1]  # GQA filter threshold
    tiles = pandas.DataFrame(p, columns=['object', 'date'])
    g = tiles.groupby('date', sort=False)

    print('{} observations; {} of {} tiles'.format(len(g), len(p), len(files)))

    return [fuser(list(obs.object)) for date, obs in g]


def show_obs(fus):
    """ Plotting (for debugging only) """
    import matplotlib.pyplot as plt
    fig, axes = plt.subplots(1, 1 + len(fus.tiles))
    for ax, w in zip(axes, [t.water for t in fus.tiles] + [fus.water]):
        ax.imshow((w & np.uint8(1))[::10, ::10])


def write(filename, data, nodata=None):
    """ Output raster to filesystem """
    import rasterio
    with rasterio.open(filename,
                       mode='w',
                       width=4000,  # should absorb into reader.cell
                       height=4000,
                       count=1,
                       dtype=data.dtype.name,
                       driver='GTIFF',
                       nodata=nodata,
                       tiled=True,
                       compress='LZW',  # balance IO volume and CPU speed
                       **reader.cell) as destination:
        destination.write(data, 1)


def main():
    """ Command-line Interface """
    import sys
    if len(sys.argv) < 3:
        print("Usage: python simple.py  OUTPUT_PREFIX  INPUT_TILE_DIRS..")  # TODO: deprecate >1 input dirs
        sys.exit(2)

    output_prefix = sys.argv[1]
    input_dirs = sys.argv[2:]

    obs = get_observations(get_filenames(*input_dirs))
    summarise_result(obs, prefix=output_prefix)


if __name__ == '__main__':
    main()
